// mismatch-fst.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Representation of mismatch counting transducer.

#ifndef KERNEL_LIB_MISMATCH_FST_H__
#define KERNEL_LIB_MISMATCH_FST_H__

#include <fst/fstlib.h>

namespace kernel {

using namespace fst;

template <class A>
void Mismatch(VectorFst<A> *fst,
              typename A::StateId k,
              typename A::StateId m,
              typename A::Label sigma) {


  for (int i = 0, s = 0; i <= m; s += (k + 1 - i), ++i) {
    for (int j = 0; j < k - i; ++j) {
      fst->AddState();
      if (FLAGS_v)
        cerr << i << " " << j << " "  << s << " " << s + j << " "
             << fst->NumStates() << endl;;
      //Add matches form s+j to s+j+1
      for (typename A::Label l = 1; l <= sigma; ++l)
        fst->AddArc(s + j, A(l, l, A::Weight::One(), s + j + 1));
      //Add mismatches from s+j to s+j+1+k-i
      if (i == m) continue;
      for (typename A::Label l1 = 1; l1 <= sigma; ++l1)
        for (typename A::Label l2 = 1; l2 <= sigma; ++l2) {
          if (l1 == l2) continue;
          fst->AddArc(s + j, A(l1, l2, A::Weight::One(), s + j + 1 + k - i));
        }
    }
    // Add final state & loops;
    fst->AddState();
    fst->SetFinal(s + k - i, A::Weight::One());
    for (typename A::Label l = 1; l <= sigma; ++l)
      fst->AddArc(s + k - i, A(l, 0, A::Weight::One(), s + k - i));
  }
  // Initial state
  fst->SetStart(0);
  for (typename A::Label l = 1; l <= sigma; ++l)
    fst->AddArc(0, A(l, 0, A::Weight::One(), 0));

  EncodeMapper<A> encoder(kEncodeLabels | kEncodeWeights, ENCODE);
  Encode(fst, &encoder);
  Minimize(fst);
  Decode(fst, encoder);
  ArcSort(fst, ILabelCompare<A>());
}



}  // namespace kernel

#endif  // KERNEL_LIB_MISMATCH_FST_H__
